import logging
import os
from contextlib import AbstractContextManager
from typing import Callable, Dict, Iterable, Iterator, Optional, Tuple, Union

from requests import Timeout
from sgqlc.operation import GraphQLErrors, Operation
from sgqlc.types import Variable

from .gql_clients import AbstractGQLClient, KeycloakAwareGQLClient, NoAuthGQLClient
from .schemas.api_schema import ConceptPropertyValueType, ConceptType, Query

logger = logging.getLogger(__name__)


class TalismanAPIAdapter(AbstractContextManager):
    def __init__(self, gql_uris: Union[str, Tuple[str, ...]]):
        self._gql_uris = (gql_uris,) if isinstance(gql_uris, str) else gql_uris
        self._gql_clients: Optional[Tuple[AbstractGQLClient, ...]] = None

    def __enter__(self):
        self._check_api()
        if os.getenv("KEYCLOAK_AUTH_URL") is None:
            self._gql_clients = tuple(NoAuthGQLClient(uri) for uri in self._gql_uris)
        else:
            self._gql_clients = tuple(KeycloakAwareGQLClient(uri) for uri in self._gql_uris)

        logger.info(f"{type(self._gql_clients[0])} will be used")

        for gql_client in self._gql_clients:
            gql_client.__enter__()
        return self

    def __exit__(self, *exc):
        for gql_client in reversed(self._gql_clients):
            gql_client.__exit__(*exc)
        self._gql_clients = None

    def _check_api(self):
        logger.info(f"used correct version of API")
        return True

    def get_base_types(self, dictionary: bool = False, regexp: bool = False, pretrained_nercmodels: bool = False
                       ) -> Tuple[Tuple[ConceptType, ...], Tuple[ConceptPropertyValueType, ...]]:
        op = Operation(Query)

        concept_types: ConceptType = op.list_concept_type
        property_value_types: ConceptPropertyValueType = op.list_concept_property_value_type

        concept_types.id()
        property_value_types.id()

        if dictionary:
            concept_types.dictionary()
            property_value_types.dictionary()

        if regexp:
            concept_types.regexp()
            property_value_types.regexp()

        if pretrained_nercmodels:
            concept_types.pretrained_nercmodels()
            property_value_types.pretrained_nercmodels()

        ret = self.gql_call(op)
        return tuple(ret.list_concept_type), tuple(ret.list_concept_property_value_type)

    def execute_easy_query(self, name: str, params: Iterable[str], variables: Optional[Dict] = None):
        op = Operation(Query)
        if variables:
            op[name](**variables).__fields__(*params)
        else:
            op[name].__fields__(*params)
        return self.gql_call(op)[name]

    def pagination_query(
            self,
            pagination_query: str, query_params: dict,
            list_query: str, list_query_builder: Callable[[Operation], None]
    ) -> Iterator:
        offset = 0
        # prepare operation
        op = Operation(Query, variables={"offset": int})
        pagination = op[pagination_query](**query_params, offset=Variable("offset"))
        pagination.total()
        list_query_builder(pagination[list_query]())

        # get first batch
        ret = self.gql_call(op, {"offset": offset})[pagination_query]
        total = ret.total
        yield from ret[list_query]

        # request until all items paginated
        offset += len(ret[list_query])
        while offset < total:  # while not
            ret = self.gql_call(op, {"offset": offset})[pagination_query]
            if not ret[list_query]:
                raise Exception("Empty query result")
            yield from ret[list_query]
            offset += len(ret[list_query])

    def _gql_call(self, gql_client: AbstractGQLClient, sgqlc_operation: Operation, variables: Optional[dict] = None,
                  raise_on_timeout: bool = True, raise_on_query: bool = False):
        try:
            return sgqlc_operation + gql_client.execute(sgqlc_operation, variables=variables)
        except Timeout as e:
            logger.error('Timeout while query processing', exc_info=e, extra={'query': sgqlc_operation})
            if raise_on_timeout:
                raise e
        except GraphQLErrors as e:
            logger.error(f"GraphQL exception '{e.errors[0]['message']}' during query processing.", exc_info=e,
                         extra={'query': sgqlc_operation})

            if raise_on_query or 'Cannot query field' not in e.errors[0]['message']:
                raise e
        except Exception as e:
            logger.error('Some exception was occured during query processing.', exc_info=e,
                         extra={'query': sgqlc_operation})
            raise e

    def gql_call(self, sgqlc_operation: Operation, variables: Optional[dict] = None, raise_on_timeout: bool = True):
        for gql_client in self._gql_clients[:-1]:
            res = self._gql_call(gql_client, sgqlc_operation, variables, raise_on_timeout)

            if res is not None:
                return res

        return self._gql_call(self._gql_clients[-1], sgqlc_operation, variables, raise_on_timeout, raise_on_query=True)



import itertools
import json
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Iterator, List, Tuple

from tp_interfaces.knowledge_base.interfaces import AbstractConceptReader, DBConcept, DBProperty, DBRelation, DBRemoteConcept


class JSONConceptReader(AbstractConceptReader):
    def __init__(self, filepath: Path):
        self._path = filepath

    def read(self) -> Iterator[DBConcept]:
        raw_objects = list(self._read_raw_objs())
        local_ids = {raw_obj["id"] for raw_obj in raw_objects if "id" in raw_obj}

        in_relations = defaultdict(list)
        out_relations = defaultdict(list)
        counter = itertools.count()

        for raw_obj in raw_objects:
            is_local = "id" in raw_obj
            id_ = raw_obj["id"] if is_local else f"DBRemoteConcept id: {next(counter)}"
            out_rels = []

            for rel in self._parse_raw_rels(raw_obj, id_):
                current_rel = rel
                if rel.target_id in local_ids:
                    in_relations[rel.target_id].append(rel)
                else:
                    current_id = f"DBRemoteConcept id: {next(counter)}"
                    current_rel = DBRelation(id_, current_id, rel.type, rel.properties)
                    yield DBRemoteConcept(current_id, rel.target_id, None, [], [current_rel], [], [])

                out_rels.append(current_rel)

            if not is_local:
                yield DBRemoteConcept(id_, raw_obj["name"], raw_obj["type"], [], [], out_rels, [])
            else:
                out_relations[id_] = out_rels

        for raw_obj in raw_objects:
            if "id" not in raw_obj:
                continue
            obj_in_rels = in_relations[raw_obj["id"]]
            obj_out_rels = out_relations[raw_obj["id"]]
            obj_prop = self._parse_raw_properties(raw_obj.get('properties', None))
            yield DBConcept(raw_obj["id"], raw_obj["name"], raw_obj["type"], raw_obj["aliases"], obj_in_rels, obj_out_rels, obj_prop)

    def _read_raw_objs(self) -> Iterator[dict]:
        with self._path.open('r', encoding='utf-8') as f:
            yield from map(json.loads, f)

    def _parse_raw_rels(self, raw_obj: dict, id_: str) -> Iterator[DBRelation]:
        return (DBRelation(id_, raw_rel["target"], raw_rel["type"], tuple(self._parse_raw_properties(raw_rel.get('properties', None))))
                for raw_rel in raw_obj["relations"])

    @staticmethod
    def _parse_raw_properties(raw_properties: tuple) -> Iterator[DBProperty]:
        if not raw_properties:
            return tuple()
        return (DBProperty(property_['name'], property_['value']) for property_ in raw_properties)


class TypesMappingFilteringReader(AbstractConceptReader):
    def __init__(
            self,
            reader: AbstractConceptReader,
            concepts_types_mapping: Dict[str, str],
            relations_types_mapping: Dict[str, Dict[str, Any]],
            properties_types_mapping: Dict[str, Dict[str, str]]
    ):
        """
        :param reader: reader to wrap
        :param concepts_types_mapping: Mapping [old concept type -> new concept type].
            If old concept type is not in mapping it will be skipped.
        :param relations_types_mapping:
            Mapping [old_rel_type] -> [source_concept_old_type, target_concept_old_type, new_rel_type, need_to_flip].
            If relation signature not in mapping it will be skipped.
            If need_to_flip flag is set for relation its source and target concepts will be flipped in new relation.
        :param properties_types_mapping:
            Mapping [old concept type -> [old property -> new property]]
        """
        self._reader = reader
        self._concepts_types_mapping = deepcopy(concepts_types_mapping)
        self._relations_types_mapping = deepcopy(relations_types_mapping)
        self._properties_types_mapping = deepcopy(properties_types_mapping)

    def read(self) -> Iterator[DBConcept]:
        id2concept: Dict[str, DBConcept] = {}
        id2type: Dict[str, str] = {}
        id2in_relations: Dict[str, List[DBRelation]] = {}
        id2out_relations: Dict[str, List[DBRelation]] = {}

        for concept in self._reader.read():
            id2concept[concept.id_] = concept
            id2type[concept.id_] = concept.type

        for id_, concept in id2concept.items():
            id2in_relations[id_], id2out_relations[id_] = self._process_relations(concept, id2type)

        for id_, concept in id2concept.items():
            if id2type[id_] not in self._concepts_types_mapping:
                continue
            prop_map = self._properties_types_mapping[id2type[id_]]
            properties = tuple(DBProperty(prop_map.get(prop.name, prop.name), prop.value)
                               for prop in concept.properties) if concept.properties else tuple()
            params = {
                "id_": concept.id_,
                "name": concept.name,
                "type_": self._concepts_types_mapping.get(id2type[id_]),
                "aliases": concept.aliases,
                "in_relations": id2in_relations[id_],
                "out_relations": id2out_relations[id_],
                "properties": properties
            }

            yield concept.create(**params)

    def _process_relations(self, concept: DBConcept, id2types: Dict[str, str]) -> Tuple[List[DBRelation], List[DBRelation]]:
        in_relations, out_relations = [], []

        for rel in concept.relations:
            if rel.type not in self._relations_types_mapping:
                continue

            rel_map = self._relations_types_mapping[rel.type]
            source_id, target_id = (rel.target_id, rel.source_id) if rel_map["flip_relation"] else (rel.source_id, rel.target_id)

            source_type = rel_map["source_type"]
            if id2types[source_id] is None:
                id2types[source_id] = rel_map["source_type"]
            elif id2types[source_id] != source_type:
                continue

            target_type = rel_map["target_type"]
            if id2types[target_id] is None:
                id2types[target_id] = target_type
            elif target_type != id2types[target_id]:
                continue

            properties = tuple(DBProperty(rel_map["properties"].get(prop.name, prop.name), prop.value)
                               for prop in rel.properties) if rel.properties else tuple()

            rel = DBRelation(source_id, target_id, rel_map["new_relation_type"], properties)

            if rel.target_id == concept.id_:
                in_relations.append(rel)
            elif rel.source_id == concept.id_:
                out_relations.append(rel)
            else:
                raise ValueError

        return in_relations, out_relations

    @classmethod
    def from_config(cls, reader: AbstractConceptReader, json_config: dict):
        concepts_map = {}
        properties_map = {}

        for old_type, new_type_with_props in json_config["concepts_types_mapping"].items():
            concepts_map[old_type] = new_type_with_props["name"]
            properties_map[old_type] = new_type_with_props.get("properties", {})

        relations_map = {}
        for obj in json_config.get("relations_types_mapping", []):
            relations_map[obj["old_relation_type"]] = {
                "source_type": obj["source_type"],
                "target_type": obj["target_type"],
                "new_relation_type": obj["new_relation_type"],
                "flip_relation": obj.get("flip_relation", False),
                "properties": obj.get("properties", {})
            }

        return cls(reader, concepts_map, relations_map, properties_map)



from abc import ABCMeta, abstractmethod
from functools import partial
from itertools import chain
from typing import Generic, Iterator, Sequence, Tuple, Type, TypeVar

from more_itertools import ichunked
from tdm.abstract.datamodel import AbstractTreeDocumentContent
from tdm.datamodel import TalismanDocument

from tp_interfaces.abstract.model import AbstractModel
from tp_interfaces.abstract.schema import ImmutableBaseModel

_DocumentContent = TypeVar('_DocumentContent', bound=AbstractTreeDocumentContent)
_Config = TypeVar('_Config', bound=ImmutableBaseModel)


class AbstractDocumentProcessor(AbstractModel, Generic[_Config, _DocumentContent], metaclass=ABCMeta):

    @abstractmethod
    def process_doc(self, document: TalismanDocument[_DocumentContent], config: _Config) -> TalismanDocument[_DocumentContent]:
        pass

    def process_docs(self, documents: Sequence[TalismanDocument[_DocumentContent]], config: _Config) \
            -> Tuple[TalismanDocument[_DocumentContent], ...]:
        return tuple(map(partial(self.process_doc, config=config), documents))

    def process_stream(self, documents: Iterator[TalismanDocument[_DocumentContent]], config: _Config, batch_size: int) \
            -> Iterator[TalismanDocument[_DocumentContent]]:
        return chain.from_iterable(map(partial(self.process_docs, config=config), map(tuple, ichunked(documents, batch_size))))

    @property
    @abstractmethod
    def config_type(self) -> Type[_Config]:
        pass



import logging
import os
import time
from typing import Optional

from keycloak import KeycloakOpenID
from sgqlc.endpoint.requests import RequestsEndpoint

from .abstract import AbstractGQLClient

logger = logging.getLogger(__name__)


class KeycloakAwareGQLClient(AbstractGQLClient):
    _TIME_OFFSET = 10  # in seconds

    def __init__(self, gql_uri: str):
        self._gql_uri = gql_uri

        self._auth_url = os.getenv("KEYCLOAK_AUTH_URL")
        self._realm = os.getenv("KEYCLOAK_REALM")
        self._client_id = os.getenv("KEYCLOAK_CLIENT_ID")
        self._client_secret = os.getenv("KEYCLOAK_CLIENT_KEY")
        self._user = os.getenv("KEYCLOAK_USER")
        self._pwd = os.getenv("KEYCLOAK_PWD")

        if any(env is None for env in [self._auth_url, self._realm, self._client_id, self._client_secret, self._user, self._pwd]):
            raise ValueError("Authorization environment values are not set")

        self._keycloak_openid: Optional[KeycloakOpenID] = None
        self._access_token: Optional[str] = None
        self._refresh_token: Optional[str] = None
        self._access_expiration_timestamp: Optional[float] = None
        self._refresh_expiration_timestamp: Optional[float] = None

        self._gql_client: Optional[RequestsEndpoint] = None

    def _ensure_session_liveness(self):
        offsetted_time = time.time() + self._TIME_OFFSET
        if self._access_expiration_timestamp is not None and offsetted_time < self._access_expiration_timestamp:
            return

        time_before_req = time.time()
        if self._refresh_expiration_timestamp is not None and offsetted_time < self._refresh_expiration_timestamp:
            logger.info("refreshing access token with refresh token")
            token_info = self._keycloak_openid.refresh_token(self._refresh_token)
        else:
            logger.info("refreshing access token with credentials")
            token_info = self._keycloak_openid.token(self._user, self._pwd)

        self._access_token = token_info['access_token']
        self._access_expiration_timestamp = time_before_req + token_info['expires_in']
        self._refresh_token = token_info['refresh_token']
        self._refresh_expiration_timestamp = time_before_req + token_info['refresh_expires_in']

        headers = {"X-Auth-Token": self._access_token, "Authorization": f"Bearer {self._access_token}"}
        self._gql_client = self._configure_gql_client(self._gql_uri, headers)

    def __enter__(self):
        self._keycloak_openid = KeycloakOpenID(self._auth_url, self._realm, self._client_id, self._client_secret)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._gql_client = None
        self._access_token, self._refresh_token = None, None
        self._access_expiration_timestamp, self._refresh_expiration_timestamp = None, None
        self._keycloak_openid, self._gql_client = None, None

    def execute(self, query, variables=None, operation_name=None, extra_headers=None, timeout=None):
        self._ensure_session_liveness()
        return self._gql_client(query, variables, operation_name, extra_headers, timeout)



__all__ = [
    'AbstractCompositeModel',
    'AbstractConfigConstructableModel', 'AbstractMultipleConfigConstructableModel',
    'AbstractModel',
    'AbstractModelWrapper'
]

from .composite import AbstractCompositeModel
from .constructable import AbstractConfigConstructableModel, AbstractMultipleConfigConstructableModel
from .model import AbstractModel
from .wrapper import AbstractModelWrapper



from abc import ABCMeta, abstractmethod
from typing import Generic, Tuple, Type, TypeVar

from tp_interfaces.abstract.model import AbstractModel
from .update import AbstractUpdate, UpdateMode

_Update = TypeVar('_Update', bound=AbstractUpdate)


class AbstractUpdatableModel(AbstractModel, Generic[_Update]):

    def update(self, update: _Update) -> None:
        if update.mode is UpdateMode.add:
            self._add(update)
        else:
            self._remove(update)

    @abstractmethod
    def info(self) -> Tuple[_Update, ...]:
        pass

    @abstractmethod
    def _add(self, update: _Update) -> None:
        pass

    @abstractmethod
    def _remove(self, update: _Update) -> None:
        pass

    @property
    @abstractmethod
    def update_type(self) -> Type[_Update]:
        pass


class UpdatableModelMixin(AbstractUpdatableModel[_Update], Generic[_Update], metaclass=ABCMeta):
    def __init__(self, model: AbstractUpdatableModel):
        self._model = model

    def info(self) -> Tuple[_Update, ...]:
        return self._model.info()

    def update(self, update: _Update) -> None:
        self._model.update(update)

    def _add(self, update: _Update) -> None:
        self._model._add(update)

    def _remove(self, update: _Update) -> None:
        self._model._remove(update)

    @property
    def update_type(self) -> Type[_Update]:
        return self._model.update_type



from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Generic, Iterable, TextIO, TypeVar

from tdm.abstract.datamodel import AbstractTreeDocumentContent
from tdm.datamodel import TalismanDocument

_Content = TypeVar('_Content', bound=AbstractTreeDocumentContent)


class AbstractSerializer(Generic[_Content], metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, doc: TalismanDocument[_Content], stream: TextIO):
        pass

    @staticmethod
    def _check_stream(stream: TextIO):
        if stream.closed or not stream.writable():
            raise Exception("stream  is closed or not writeable")


class AbstractPathSerializer(Generic[_Content], metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, docs: Iterable[TalismanDocument[_Content]], path: Path):
        pass
